

#include <windows.h>
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include <limits.h>

#include <io.h>
#include <string>
#include <vector>
#include <map>

#include "filter.h"
#include "scalex.h"

extern void InitLUTs();
extern void UninitLUTs();
extern void hq2x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );
extern void hq3x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );
extern void hq4x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );


enum FilterMode {
	N2X = 1,
	SCALE2X,
	HQ2X,
	N3X,
	SCALE3X,
	HQ3X,
	N4X,
	SCALE4X,
	HQ4X
};

// [min, max] ͈̔͂ɖOa
template <class T>
T saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

template <>
int saturate(int v, int min, int max) {
    v -= min;
    v &= ~(v >> 31);
    v += min;

    v -= max;
    v &=  (v >> 31);
    v += max;
    
    return v;
}


static inline int modulo(int a, int b) {
	int m = a % b;

	if (m >= 0) {
		return m;
	} else {
		return m + b;
	}
}



void CopyBufToRGB565(
		PIXEL_YC *buf, WORD *img,
		int x0, int y0, int w, int h, int x1, int y1, int mw0, int mh0, int mw1, int mh1
){
	int src,dst,ix0,iy0,ix1,iy1;

	for(iy0 = y0, iy1 = y1; iy0 < y0 + h && iy0 < mh0 && iy1 < mh1; iy0++,iy1++) {
		src = x0 + iy0 * mw0;
		dst = x1 + iy1 * mw1;

		for(ix0 = x0, ix1 = x1; ix0 < x0 + w && ix0 < mw0 && ix1 < mw1; ix0++,ix1++) {
			int y  = buf[src].y;
			int cb = buf[src].cb;
			int cr = buf[src].cr;

			int r = saturate(( 255*y + ( ((22881*cr>>16)+3) <<10) ), 0, 256*4096-1);
			int g = saturate(( 255*y + ( ((-5616*cb>>16)+(-11655*cr>>16)+3) <<10) ), 0, 256*4096-1);
			int b = saturate(( 255*y + ( ((28919*cb>>16)+3) <<10) ), 0, 256*4096-1);

			int r2 = r / (8*4096);
			int g2 = g / (4*4096);
			int b2 = b / (8*4096);

			WORD col = (r2 << 11) + (g2 << 5) + b2;
			img[dst] = col;

			src++;
			dst++;
		}
	}
}


void CopyRGB8888ToBuf(
		DWORD *img, PIXEL_YC *buf, 
		int x0, int y0, int w, int h, int x1, int y1, int mw0, int mh0, int mw1, int mh1
){
	int src,dst,ix0,iy0,ix1,iy1;
	PIXEL_YC yc;

	for(iy0 = y0, iy1 = y1; iy0 < y0 + h && iy0 < mh0 && iy1 < mh1; iy0++,iy1++) {
		src = x0 + iy0 * mw0;
		dst = x1 + iy1 * mw1;

		for(ix0 = x0, ix1 = x1; ix0 < x0 + w && ix0 < mw0 && ix1 < mw1; ix0++,ix1++) {
			DWORD col = img[src];
			int r = ((col & 0x00FF0000) >> 16) + 4;
			int g = ((col & 0x0000FF00) >> 8) + 2;
			int b = (col & 0x000000FF) + 4;

			yc.y =  ((4918*r+354)>>10) + ((9655*g+585)>>10) + ((1875*b+523)>>10);
			yc.cb = ((-2775*r+240)>>10) + ((-5449*g+515)>>10) + (( 8224*b+256)>>10);
			yc.cr = ((8224*r+256)>>10) + ((-6887*g+110)>>10) + ((-1337*b+646)>>10);

			buf[dst] = yc;

			src++;
			dst++;
		}
	}
}



static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;
static std::vector<int> g_bef_check;
static int g_video_rate  = 30;
static int g_video_scale = 1;

static int getTrackVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_trackMap.find(str);

	if (itr == g_trackMap.end()) {
		sprintf_s(errmsg, "track %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	int val = fp->track[itr->second];
	val = std::max<int>(val, fp->track_s[itr->second]);
	val = std::min<int>(val, fp->track_e[itr->second]);

	return val;
}

static int getCheckVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_checkMap.find(str);

	if (itr == g_checkMap.end()) {
		sprintf_s(errmsg, "check %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->check[itr->second];
}

void initFilter(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}
	g_bef_check.resize(fdp->check_n, 0);

	InitLUTs();
}

void updateFilter(FILTER *fp) {
	int imode = g_trackMap["[h"];

	int mode = getTrackVal(fp, "[h");
	char *p_name = "";

	switch(mode) {
	case N2X:
		p_name = "2x";
		break;
	case N3X:
		p_name = "3x";
		break;
	case N4X:
		p_name = "4x";
		break;
	case SCALE2X:
		p_name = "scale2x";
		break;
	case SCALE3X:
		p_name = "scale3x";
		break;
	case SCALE4X:
		p_name = "scale4x";
		break;
	case HQ2X:
		p_name = "hq2x";
		break;
	case HQ3X:
		p_name = "hq3x";
		break;
	case HQ4X:
		p_name = "hq4x";
		break;
	}
	strcpy_s(fp->track_name[imode], 256, p_name);

	fp->exfunc->filter_window_update(fp);
	return;
}


void exitFilter() {
	UninitLUTs();
}

// e덷
static int g_Ymgn  = 0;
static int g_Cbmgn = 0;
static int g_Crmgn = 0;

inline bool isEqual(const PIXEL_YC& pix1, const PIXEL_YC& pix2) {
	int dy  = abs(pix1.y  - pix2.y);
	int dcb = abs(pix1.cb - pix2.cb);
	int dcr = abs(pix1.cr - pix2.cr);

	if (dy <= g_Ymgn && dcb <= g_Cbmgn && dcr <= g_Crmgn) {
		return true;
	} else {
		return false;
	}
}

void execNx(PIXEL_YC *ibuf, PIXEL_YC *obuf,
			FILTER_PROC_INFO *fpip, int N) {

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	for (int iy = 0; iy < h; iy++) {
		if (iy*N >= mh) break;

		for (int ix = 0; ix < w; ix++) {
			if (ix*N >= mw) break;

			PIXEL_YC E = ibuf[iy*mw + ix];
			for (int iy2 = 0; iy2 < N && iy*N+iy2 < mh; iy2++) {
				for (int ix2 = 0; ix2 < N && ix*N+ix2 < mw; ix2++) {
					obuf[(iy*N+iy2)*mw + (ix*N+ix2)] = E;
				}
			}
		}
	}
	
	int aw = std::min<int>(fpip->w * N, mw);
	int ah = std::min<int>(fpip->h * N, mh);
	fpip->w = aw;
	fpip->h = ah;
}


inline void calcScale2x(
				                   const PIXEL_YC& B,
				const PIXEL_YC& D, const PIXEL_YC& E, const PIXEL_YC& F,
				                   const PIXEL_YC& H,
				PIXEL_YC* pE0, PIXEL_YC* pE1, 
				PIXEL_YC* pE2, PIXEL_YC* pE3) {

    if (!isEqual(B, H) && !isEqual(D, F)) {
    	*pE0 = isEqual(D, B) ? B : E;
    	*pE1 = isEqual(B, F) ? F : E;
    	*pE2 = isEqual(D, H) ? D : E;
    	*pE3 = isEqual(H, F) ? H : E;
    } else {
    	*pE0 = E;
    	*pE1 = E;
    	*pE2 = E;
    	*pE3 = E;
    }
}

void execScale2x(PIXEL_YC *ibuf, PIXEL_YC *obuf,
				 FILTER_PROC_INFO *fpip) {

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	for (int iy = 0; iy < h; iy++) {
		if (iy*2 >= mh) break;

		int iy0 = std::max<int>(iy-1, 0);
		int iy1 = iy;
		int iy2 = std::min<int>(iy+1, mh-1);

		for (int ix = 0; ix < w; ix++) {
			if (ix*2 >= mw) break;

			int ix0 = std::max<int>(ix-1, 0);
			int ix1 = ix;
			int ix2 = std::min<int>(ix+1, mw-1);

			PIXEL_YC B = ibuf[iy0*mw + ix1];
			PIXEL_YC D = ibuf[iy1*mw + ix0];
			PIXEL_YC E = ibuf[iy1*mw + ix1];
			PIXEL_YC F = ibuf[iy1*mw + ix2];
			PIXEL_YC H = ibuf[iy2*mw + ix1];
			PIXEL_YC E0;
			PIXEL_YC E1;
			PIXEL_YC E2;
			PIXEL_YC E3;
				
			calcScale2x(B,D,E,F,H, &E0,&E1,&E2,&E3);

			obuf[(iy*2  )*mw + (ix*2  )] = E0;
			if (ix*2+1 < mw) {
				obuf[(iy*2  )*mw + (ix*2+1)] = E1;
			}
			if (iy*2+1 < mh) {
				obuf[(iy*2+1)*mw + (ix*2  )] = E2;
				if (ix*2+1 < mw) {
					obuf[(iy*2+1)*mw + (ix*2+1)] = E3;
				}
			}
		}
	}
	
	int aw = std::min<int>(fpip->w * 2, mw);
	int ah = std::min<int>(fpip->h * 2, mh);
	fpip->w = aw;
	fpip->h = ah;
}

inline void calcScale3x(
				const PIXEL_YC& A, const PIXEL_YC& B, const PIXEL_YC& C,
				const PIXEL_YC& D, const PIXEL_YC& E, const PIXEL_YC& F,
				const PIXEL_YC& G, const PIXEL_YC& H, const PIXEL_YC& I,
				PIXEL_YC* pE0, PIXEL_YC* pE1, PIXEL_YC* pE2, 
				PIXEL_YC* pE3, PIXEL_YC* pE4, PIXEL_YC* pE5, 
				PIXEL_YC* pE6, PIXEL_YC* pE7, PIXEL_YC* pE8
				) {
	if (!isEqual(B,H) && !isEqual(D,F)) {
    	*pE0 = isEqual(D,B) ? D : E;
    	*pE1 = (isEqual(D,B) && !isEqual(E,C)) || (isEqual(B,F) && !isEqual(E,A)) ? B : E;
    	*pE2 = isEqual(B,F) ? F : E;
    	*pE3 = (isEqual(D,B) && !isEqual(E,G)) || (isEqual(D,H) && !isEqual(E,A)) ? D : E;
    	*pE4 = E;
    	*pE5 = (isEqual(B,F) && !isEqual(E,I)) || (isEqual(H,F) && !isEqual(E,C)) ? F : E;
    	*pE6 = isEqual(D,H) ? D : E;
    	*pE7 = (isEqual(D,H) && !isEqual(E,I)) || (isEqual(H,F) && !isEqual(E,G)) ? H : E;
    	*pE8 = isEqual(H,F) ? F : E;
    } else {
    	*pE0 = E;
    	*pE1 = E;
    	*pE2 = E;
    	*pE3 = E;
    	*pE4 = E;
    	*pE5 = E;
    	*pE6 = E;
    	*pE7 = E;
    	*pE8 = E;
    }
}

void execScale3x(PIXEL_YC *ibuf, PIXEL_YC *obuf,
				 FILTER_PROC_INFO *fpip) {

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	for (int iy = 0; iy < h; iy++) {
		if (iy*3 >= mh) break;

		int iy0 = std::max<int>(iy-1, 0);
		int iy1 = iy;
		int iy2 = std::min<int>(iy+1, mh-1);

		for (int ix = 0; ix < w; ix++) {
			if (ix*3 >= mw) break;

			int ix0 = std::max<int>(ix-1, 0);
			int ix1 = ix;
			int ix2 = std::min<int>(ix+1, mw-1);

			PIXEL_YC A = ibuf[iy0*mw + ix0];
			PIXEL_YC B = ibuf[iy0*mw + ix1];
			PIXEL_YC C = ibuf[iy0*mw + ix2];
			PIXEL_YC D = ibuf[iy1*mw + ix0];
			PIXEL_YC E = ibuf[iy1*mw + ix1];
			PIXEL_YC F = ibuf[iy1*mw + ix2];
			PIXEL_YC G = ibuf[iy2*mw + ix0];
			PIXEL_YC H = ibuf[iy2*mw + ix1];
			PIXEL_YC I = ibuf[iy2*mw + ix2];
			PIXEL_YC E0, E1, E2;
			PIXEL_YC E3, E4, E5;
			PIXEL_YC E6, E7, E8;
				
			calcScale3x(A,B,C,D,E,F,G,H,I, 
						&E0,&E1,&E2,&E3,&E4,&E5,&E6,&E7,&E8);

			obuf[(iy*3  )*mw + (ix*3  )] = E0;
			if (ix*3+1 < mw) {
				obuf[(iy*3  )*mw + (ix*3+1)] = E1;
			}
			if (ix*3+2 < mw) {
				obuf[(iy*3  )*mw + (ix*3+2)] = E2;
			}

			if (iy*3+1 < mh) {
				obuf[(iy*3+1)*mw + (ix*3  )] = E3;
				if (ix*3+1 < mw) {
					obuf[(iy*3+1)*mw + (ix*3+1)] = E4;
				}
				if (ix*3+2 < mw) {
					obuf[(iy*3+1)*mw + (ix*3+2)] = E5;
				}
			}

			if (iy*3+2 < mh) {
				obuf[(iy*3+2)*mw + (ix*3  )] = E6;
				if (ix*3+1 < mw) {
					obuf[(iy*3+2)*mw + (ix*3+1)] = E7;
				}
				if (ix*3+2 < mw) {
					obuf[(iy*3+2)*mw + (ix*3+2)] = E8;
				}
			}
		}
	}
	
	int aw = std::min<int>(fpip->w * 3, mw);
	int ah = std::min<int>(fpip->h * 3, mh);
	fpip->w = aw;
	fpip->h = ah;
}



void execHqNx(PIXEL_YC *ibuf, PIXEL_YC *obuf,
			  FILTER_PROC_INFO *fpip, int N) {

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	WORD  *pRGB565  = new WORD[w*h];
	DWORD *pRGB8888 = new DWORD[(w*N)*(h*N)];

	CopyBufToRGB565(ibuf, pRGB565, 0,0,w,h,0,0, mw,mh, w,h);

	switch(N) {
	case 2:
		hq2x_32((unsigned char*)pRGB565, (unsigned char*)pRGB8888, w, h, (w*N)*4);
		break;
	case 3:
		hq3x_32((unsigned char*)pRGB565, (unsigned char*)pRGB8888, w, h, (w*N)*4);
		break;
	case 4:
		hq4x_32((unsigned char*)pRGB565, (unsigned char*)pRGB8888, w, h, (w*N)*4);
		break;
	}

	CopyRGB8888ToBuf(pRGB8888, obuf, 0,0,w*N,h*N,0,0, w*N,h*N, mw,mh);
	
	delete[] pRGB565;
	delete[] pRGB8888;
	
	int aw = std::min<int>(fpip->w * N, mw);
	int ah = std::min<int>(fpip->h * N, mh);
	fpip->w = aw;
	fpip->h = ah;
}


BOOL execFilter(FILTER *fp,FILTER_PROC_INFO *fpip) {
	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	g_Ymgn  = getTrackVal(fp, "eY");
	g_Cbmgn = getTrackVal(fp, "eCb");
	g_Crmgn = getTrackVal(fp, "eCr");

	int mode = getTrackVal(fp, "[h");

	switch(mode) {
	case N2X:
		execNx(buffer0, buffer1, fpip, 2);
		break;

	case N3X:
		execNx(buffer0, buffer1, fpip, 3);
		break;

	case N4X:
		execNx(buffer0, buffer1, fpip, 4);
		break;

	case SCALE2X:
		execScale2x(buffer0, buffer1, fpip);
		break;

	case SCALE3X:
		execScale3x(buffer0, buffer1, fpip);
		break;

	case SCALE4X: {
		PIXEL_YC *tmpbuf = (PIXEL_YC*)malloc((fpip->max_w * fpip->max_h) * sizeof(PIXEL_YC));
		if (tmpbuf == NULL) {
			return FALSE;
		}
		execScale2x(buffer0, tmpbuf, fpip);
		execScale2x(tmpbuf, buffer1, fpip);
		free(tmpbuf);
		break; }

	case HQ2X:
		execHqNx(buffer0, buffer1, fpip, 2);
		break;
	case HQ3X:
		execHqNx(buffer0, buffer1, fpip, 3);
		break;
	case HQ4X:
		execHqNx(buffer0, buffer1, fpip, 4);
		break;
	}

	fpip->ycp_edit = buffer1;
	fpip->ycp_temp = buffer0;

	return TRUE;
}
